#include <stdio.h>
#include <string.h>
#include "bsp.h"
#include "tty.h"
#include "flash.h"
#include "power.h"
#include "led.h"
#include "uip/uip.h"
#include "uip/timer.h"
#include "packet.h"
#include "spi.h"
#include "htonl.h"

static struct uip_udp_conn *pm_send_conn;
static struct uip_udp_conn *pm_recv_conn;
static struct timer pm_timer;

pm_appcall ()
{
  cprintf("pm_appcall\n");
}

static int init_needed;
static long long aenergy[16];
static long lenergy[16];
static long max_energy = 1;

static void
gather_energy ()
{
  int c;
  int nr = 0;
  long ng = 0;

  spi_power ();
  for (c=0; c<16; c++)
    {
      if (power_getreg (c, DIEREV) != 0xff)
	{
	  aenergy[c] += power_getreg (c, RAENERGY) * flash_data.power[c].scale;
	  lenergy[c] = power_getreg (c, LAENERGY) * flash_data.power[c].scale;
	  if (lenergy[c] < 0)
	    nr = 255;
	  else
	    ng += lenergy[c];
	}
      else
	{
	  lenergy[c] = 0;
	}
    }
  if (init_needed)
    {
      memset (aenergy, 0, sizeof(aenergy));
      memset (lenergy, 0, sizeof(lenergy));
      init_needed = 0;
    }
  if (max_energy < ng + 1)
    max_energy = ng + 1;
  green = ng * 256 / max_energy;
  if (ng && !green)
    green = 1;
  red = nr;
}

static int
incdec (int amount, int channel, int reg, int val, int min, int max)
{
  val += amount;
  if (val < min)
    val = min;
  if (val > max)
    val = max;
  power_putreg (channel, reg, val);
  return val;
}

void
calshow (int channel)
{
  word linecyc = power_getreg (channel, LINCYC);
  int wgain = power_getreg (channel, WGAIN);
  int wdiv = power_getreg (channel, WDIV);
  cprintf("LINECYC %d WDIV %d WGAIN %d  ", linecyc, wdiv, wgain);
}

void
calibrate (int channel, DF_Power *power)
{
  static long old_aenergy = 0;
  static int old_atic = 0;
  long aenergy;
  long laenergy;
  int atic;
  double aps;
  long d;
  int deltaa = 10;

  cprintf("calibrating channel %d: l/L changes LINECYC, d/D WDIV, g/G/t/T/h/H WGAIN, ESC to exit.\n", channel);
  calshow (channel);
  tty_puts ("\n");

  while (1)
    {
      spi_power ();
      atic = tics;
      aenergy = power_getreg (channel, AENERGY) * flash_data.power[channel].scale;

      if (aenergy - old_aenergy >= deltaa && atic != old_atic)
	{
	  laenergy = power_getreg (channel, LAENERGY) * flash_data.power[channel].scale;
	  aps = ((double)aenergy - old_aenergy) / ((int)atic - (int)old_atic);
	  d = aps * 1000000.0;
	  cprintf("\rAE %l  LAE %l  APS %l  ", aenergy, laenergy, d);

	  if (deltaa > 2 && (atic-old_atic) > 8000)
	    deltaa /= 2;
	  else if (deltaa > 1 && (atic-old_atic) > 4000)
	    deltaa --;
	  if (deltaa < 100 && (atic-old_atic) < 1000)
	    deltaa *= 2;
	  if (deltaa < 200 && (atic-old_atic) < 3000)
	    deltaa ++;

	  old_aenergy = aenergy;
	  old_atic = atic;
	}
      if (tty_ready ())
	{
	  char ch = tty_readc ();
	  switch (ch)
	    {
	    case 'l':
	      power->linecyc = incdec (-1, channel, LINCYC, power->linecyc, 1, 0x7fff);
	      break;
	    case 'L':
	      power->linecyc = incdec (1, channel, LINCYC, power->linecyc, 2, 0x7fff);
	      break;
	    case 'd':
	      power->wdiv = incdec (-1, channel, WDIV, power->wdiv, 1, 0xff);
	      break;
	    case 'D':
	      power->wdiv = incdec (1, channel, WDIV, power->wdiv, 1, 0xff);
	      break;
	    case 'g':
	      power->wgain = incdec (-1, channel, WGAIN, power->wgain, -2048, 2047);
	      break;
	    case 'G':
	      power->wgain = incdec (1, channel, WGAIN, power->wgain, -2048, 2047);
	      break;
	    case 't':
	      power->wgain = incdec (-10, channel, WGAIN, power->wgain, -2048, 2047);
	      break;
	    case 'T':
	      power->wgain = incdec (10, channel, WGAIN, power->wgain, -2048, 2047);
	      break;
	    case 'h':
	      power->wgain = incdec (-100, channel, WGAIN, power->wgain, -2048, 2047);
	      break;
	    case 'H':
	      power->wgain = incdec (100, channel, WGAIN, power->wgain, -2048, 2047);
	      break;
	    case 27:
	      cprintf("\n\n");
	      return;
	    }

	  tty_puts ("\n");
	  calshow (channel);
	  tty_puts ("\033[A");
	}
    }  
}

pm_set_timer ()
{
  timer_set (&pm_timer, CLOCK_CONF_SECOND * flash_data.header.poll_sec);
}

pm_udp_appcall ()
{
  if (uip_flags == UIP_POLL
      && timer_expired(&pm_timer))
    {
      PacketStruct *ps = (PacketStruct *)uip_appdata;
      int c;

      timer_reset (&pm_timer);

      memcpy (ps->magic, UDP_PACKET_MAGIC, 8);
      ps->uptime = htonl (uptime_seconds);
      ps->id = flash_data.header.id;

      gather_energy ();

      for (c=0; c<16; c++)
	{
	  ps->aenergy[c] = htonll (aenergy[c]);
	  ps->laenergy[c] = htonl (lenergy[c]);
	}
      spi_ethernet ();
      uip_udp_send (sizeof (*ps));
      return;
    }
  if (uip_flags == UIP_NEWDATA)
    {
      int i;
      cprintf("new data len %d\n", uip_len);
      for (i=0; i<uip_len; i++)
	cprintf(" %b", ((byte *)uip_appdata)[i]);
      cprintf("\n");
      return;
    }
}

throb_leds ()
{
  int i;
#define PAUSE wait_ms(1)
  for (i=0; i<256; i++)
    {
      red = green = blue = yellow = i;
      PAUSE;
    }
  for (i=0; i<256; i++)
    {
      red = green = blue = yellow = 255-i;
      PAUSE;
    }
#undef PAUSE
}

void
uip_log (char *msg)
{
  cprintf("UIP: %s\n", msg);
}

main()
{
  int vrms, irms, laenergy, aenergy;
  uip_ipaddr_t remote_ip;

  throb_leds();
  blue = 255;

  tty_init ();

  cprintf("initializing flash, ");
  flash_init ();
  cprintf("spi, ");
  spi_reset ();
  cprintf("power, ");
  power_init ();
  cprintf("done.\n");

  ether_init_uip ();
  uip_ipaddr (&remote_ip, 255, 255, 255, 255);
  pm_send_conn = uip_udp_new (&remote_ip, HTONS (PM_PORT));
  uip_udp_bind (pm_send_conn, HTONS(PM_PORT));

  pm_recv_conn = uip_udp_listen (HTONS (PM_CPORT));

  if (flash_data.header.magic != FLASH_MAGIC)
    {
      cprintf("\033[31mFlash data uninitialized!  Initializing... (don't forget to save it)\033[0m\n");
      flash_data_init();
      yellow = 255;
      flash_cmdline();
      yellow = 0;
    }

  pm_set_timer ();

  init_needed = 1;

  blue = 32;
  while (1)
    {
      ether_poll_uip ();
      if (tty_ready ())
	{
	  yellow = 255;
	  tty_readc ();
	  flash_cmdline ();
	  timer_restart (&pm_timer);
	  yellow = 0;
	}
    }
}
